package work.xuye.source.handler;

import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.cloud.stream.function.StreamBridge;
import org.springframework.expression.spel.standard.SpelExpressionParser;
import org.springframework.messaging.Message;
import org.springframework.messaging.support.MessageBuilder;
import org.springframework.stereotype.Component;
import org.springframework.util.DigestUtils;
import org.springframework.util.ObjectUtils;
import org.springframework.util.StopWatch;
import work.xuye.common.constant.*;
import work.xuye.common.db.entity.Task;
import work.xuye.common.db.service.TaskManager;
import work.xuye.common.dto.HttpRequestParams;
import work.xuye.common.dto.HttpRes;
import work.xuye.common.dto.TaskVO;
import work.xuye.common.enums.ProcessMode;
import work.xuye.common.enums.ResourceStatus;
import work.xuye.common.service.UrlMD5Service;
import work.xuye.common.utils.DebugUtil;
import work.xuye.common.utils.ID;
import work.xuye.source.request.RequestClient;

import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;

/**
 * @author xuye
 * @since 2023/2/19 16:37
 **/
@Slf4j
@Component
@RequiredArgsConstructor
public class SourceHandler {

    private final SpelExpressionParser parser;
    private final TaskManager taskManager;
    private final RequestClient requestClient;
    private final UrlMD5Service urlMD5Service;
    private final StreamBridge streamBridge;


    public Message<HttpRes> handle(Message<HttpRequestParams> message) {
        TaskVO taskVO = taskManager.getTaskInfoByTaskId((Integer) message.getHeaders().get(MessageConstants.TASK_ID));
        String taskName = taskVO.getTask().getName();
        if (DebugUtil.isSkip(taskVO)) {
            return null;
        }
        HttpRes res = null;
        HttpRequestParams requestParams = message.getPayload();
        StopWatch stopWatch = new StopWatch();
        try {
            stopWatch.start();
            res = this.request(requestParams);
            stopWatch.stop();
            String md5Digest = DigestUtils.md5DigestAsHex(res.getBody().getBytes());
            ResourceStatus resourceStatus = urlMD5Service.getResourceStatus(requestParams.getUrl(), md5Digest);
            urlMD5Service.put(requestParams.getUrl(), md5Digest);
            res.setResourceStatus(resourceStatus);
        } catch (Exception e) {
            log.error("[{}] request error, request params:[{}], error message:[{}]", taskName, requestParams, e.getMessage());
            return null;
        }
        if (!res.getStatus().is2xxSuccessful()) {
            log.warn("[{}] response status code is not 2xx, request params: [{}], response: [{}]", taskName, requestParams, res);
            throw new RuntimeException("[" + taskName + "]" + "unexpected status code: " + res.getStatus());
        }
        ProcessMode processMode = taskManager
                .getTaskInfoByTaskId(
                        (Integer) message.getHeaders().get(MessageConstants.TASK_ID))
                .getTask()
                .getProcessMode();
        this.log(message, res, processMode, stopWatch.getLastTaskTimeMillis(), taskVO.getTask());
        if (res.getResourceStatus().equals(ResourceStatus.UNCHANGED) && !ProcessMode.DEBUG.equals(processMode)) {
            return null;
        }

        if (ProcessMode.WATCH.equals(processMode)) {
            Message<String> watchMessage = MessageBuilder
                    .withPayload(res.getBody())
                    .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.SOURCE_WATCH)
                    .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                    .setHeader(MessageConstants.SOURCE_TRACE_ID, ID.generate())
                    .setHeader(MessageConstants.TASK_ID, message.getHeaders().get(MessageConstants.TASK_ID))
                    .setHeader(MessageConstants.TASK_NAME, message.getHeaders().get(MessageConstants.TASK_NAME))
                    .setHeader(SnapshotConstants.URL, requestParams.getUrl())
                    .setHeader(SnapshotConstants.RESOURCE_STATUS, res.getResourceStatus())
                    .setHeader(SnapshotConstants.STATUS, res.getStatus())
                    .build();

            boolean send = streamBridge.send(taskName + BindingConstants.SNAPSHOT_OUT_POSTFIX, watchMessage);
            if (!send) {
                throw new RuntimeException("send message failed");
            }
            return null;
        }
        return MessageBuilder
                .withPayload(res)
                .setHeader(MessageConstants.CURRENT_STAGE, StageConstants.SOURCE_WATCH)
                .setHeader(MessageConstants.TASK_TRACE_ID, message.getHeaders().get(MessageConstants.TASK_TRACE_ID))
                .setHeader(MessageConstants.SOURCE_TRACE_ID, ID.generate())
                .setHeader(MessageConstants.TASK_ID, message.getHeaders().get(MessageConstants.TASK_ID))
                .setHeader(MessageConstants.TASK_NAME, message.getHeaders().get(MessageConstants.TASK_NAME))
                .setHeader(MessageConstants.SEED_URL, requestParams.getUrl())
                .build();
    }

    private void log(Message<HttpRequestParams> message, HttpRes res, ProcessMode processMode, long ms, Task task) {
        HttpRequestParams req = message.getPayload();
        String emoji = "";
        if (processMode.equals(ProcessMode.NORMAL)) {
            emoji = "";
        } else if (processMode.equals(ProcessMode.DEBUG)) {
            emoji = "🤖";
        } else if (processMode.equals(ProcessMode.WATCH)) {
            emoji = "👀";
        }
        if (res.getStatus().is2xxSuccessful()) {
            log.info("[{}-{}] [{}{} {}] [{} {} {}ms] {} ", task.getId(), task.getName(),
                    res.getResourceStatus().getEmoji(), emoji,
                    res.getResourceStatus().getName(), req.getMethod(), res.getStatus(), ms, req.getUrl());
        } else {
            log.warn("@@ method: [{}], status: [{}], url: [{}], request: [{}], response: [{}]",
                    req.getMethod(), res.getStatus(), req.getUrl(), req.getBody(), res.getBody());
        }
    }

    public HttpRes request(HttpRequestParams req) {
        this.preHandleUrl(req);
        this.preHandleBody(req);
        return requestClient.execute(req);
    }

    private void preHandleBody(HttpRequestParams request) {
        Map<String, Object> body = request.getBody();
        if (ObjectUtils.isEmpty(body)) {
            return;
        }
        for (String key : body.keySet()) {
            Object value = body.get(key);
            Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(value.toString());
            if (matcher.find()) {
                String placeholder = matcher.group();
                String expr = request.getPlaceholderExpressions().get(placeholder);
                Object result = parser.parseExpression(expr).getValue();
                body.put(key, result);
            }
        }
        request.setBody(body);
    }

    private void preHandleUrl(HttpRequestParams request) {
        String url = request.getUrl();
        Matcher matcher = RegexConstants.PLACEHOLDER_PATTERN.matcher(url);
        Map<String, Object> patternMap = new HashMap<>();
        while (matcher.find()) {
            String placeholder = matcher.group();
            patternMap.put(placeholder, null);
        }

        if (ObjectUtils.isEmpty(patternMap)) {
            return;
        }
        Map<String, String> placeholderExpressions = request.getPlaceholderExpressions();

        for (String key : patternMap.keySet()) {
            String expr = placeholderExpressions.get(key);
            Object value = parser.parseExpression(expr).getValue();
            patternMap.put(key, value);
        }
        for (String key : patternMap.keySet()) {
            url = url.replace(key, patternMap.get(key).toString());
        }
        request.setUrl(url);
    }


}



